
/*
 *  LOAD_MAIL.C
 *
 *  file-io routines to scan the mail file and load required information.
 *
 *
 *  Global Routines:    HOLD_LOAD()         hold on loading mail after change
 *                      NOHOLD_LOAD()       hold off.. load if changes
 *                      LOAD_CHANGES()      reload mail if changed 
 *                      LOAD_MAIL()         load/reload mail
 *                      SAVE_FILE()         save mail items back to spool
 *                      CHECK_NEW_MAIL()    check for new mail
 *                      WRITE_FILE()        append mail items to a file
 *                      GET_EXTRA_OVR()     ret index of Field (create if not)
 *                      ADD_EXTRA()         add another field (reloads mail)
 *                      DELETE_EXTRA()      delete a field
 *                      GET_EXTRA()         ret index of Field, or error
 *                      M_SELECT()          select on current message list
 *
 *
 *  Static Routines:    LOAD_HASH()         load hash table from fields list
 *                      FREE_ENTRY()        unload EVERYTHING
 *                      FREE_TABLE()        unload all Fields table
 *                      LOAD_FILE()         raw file loading/counting
 *                      
 *
 */

#include "dmail.h"

Prototype int load_mail(int at, int rescan);
Prototype void hold_load(void);
Prototype void nohold_load(void);
Prototype int write_file(char *file, int modes, int mark, int notmark);
Prototype int write_prog(char *prog, int modes, int mark, int notmark);
Prototype int m_select(char *sav[], int mode, int tag);

Prototype int initial_load_mail(void);

Prototype void free_entry(void);
Prototype int free_table(int at);
Prototype int save_file(int reload, int mark, int notmark);
Prototype void check_new_mail(void);
Prototype int get_extra_ovr(char *str);
Prototype int add_extra(char *str);
Prototype int get_extra(char *str);
Prototype int position_at_from(int pos);
Prototype int position_next(int *ppos);
Prototype int data_line_length(const char *p, int isHeader);

Prototype struct FIND Find[MAXTYPE + 1];

Prototype const char *FileData;
Prototype int	FileSize;
Prototype int	FileFd;

void load_changes(void);
void do_flock(int fd, int stat);
int load_file(int at, int pos);

#define NOHOLD  0
#define HOLD    1

#define NO_BASE     0
#define NO_FIELDS   1
#define ENTRY_OK    2

struct FIND Find[MAXTYPE + 1] = {
        { "From:"   , 5, 1, 0 },
        { "To:"     , 3, 1, 0 },
/*
        { "Cc:"     , 3, 1, 0 },
        { "Newsgroups:" , 11, 1, 0 },
        { "Date:"   , 5, 1, 0 },
*/
        { "Subject:", 8, 1, 0 },
        { "References:", 11, 1, 0 },
        { "Message-ID:", 11, 1, 0 },
};

int	changed;
int	load_hold;
const char *FileData;
int	FileSize;
int	FileFd = -1;

char *quo_quo = "";

void
hold_load(void)
{
    load_hold = 1;
}

void
nohold_load(void)
{
    load_hold = 0;
    load_changes();
}

void
load_changes(void)
{
    if (changed && !load_hold)
        load_mail(Entries, 1);
}

int
initial_load_mail(void)
{
    if (load_mail(0, 0) < 0)
        return (-1);
    return ((Entries) ? 1 : -1);
}

int
load_mail(int at, int rescan)
{
    int pos;
    int fd;
    int fileSize;
    int oldEntries = Entries;
    const char *fileData;

    if (No_load_mail)
        return (-1);

    push_break();

    /*
     * Chop entry list to index 'at'.
     */

    pos = free_table(at);

    /*
     * open, lock, and map the file
     */

    if ((fd = open(mail_file, O_RDWR)) < 0) {
	pop_break();
	return(-1);
    }

    do_flock (fd, LOCK_EX);

    {
	struct stat st;

	if (fstat(fd, &st) < 0) {
	    close(fd);
	    pop_break();
	    return(-1);
	}
	if ((fileSize = st.st_size) < pos) {
	    printf("Active mail file got smaller, was ripped out from under us!\n");
	    close(fd);
	    pop_break();
	    return(-1);
	}
	fileData = mmap((caddr_t)0, fileSize, PROT_READ, MAP_SHARED, fd, 0);
	if (fileData == (void *)-1) {
	    printf("Unable to mmap mail file\n");
	    close(fd);
	    pop_break();
	    return(-1);
	}
    }

    /*
     * Replace current map with new map and renumber array
     */

    if (FileFd >= 0) {
	close(FileFd);
	FileFd = -1;
    }

    if (FileData) {
	munmap((caddr_t)FileData, FileSize);
	/* do not null out, we have to readjust pointers */
    }

    {
	int i;

	for (i = 0; i < Entries; ++i) {
	    struct ENTRY *en = &Entry[i];
	    int j;

	    if (en->from >= FileData && en->from < FileData + FileSize)
		en->from = (en->from - FileData) + fileData;

	    for (j = 0; Find[j].search != NULL; ++j) {
		const char *p = en->fields[j];

		if (p >= FileData && p < FileData + FileSize)
		    p = (p - FileData) + fileData;
		en->fields[j] = p;
	    }
	}
	FileData = fileData;
	FileSize = fileSize;
	FileFd = fd;
    }

    /*
     * Scan in new mail
     */

    if (rescan)
	load_file(0, 0);
    else
	load_file(at, pos);

    if (oldEntries != Entries) {
	if (!lmessage_overide) {
	    if (oldEntries == 0)
		printf("%d Items Loaded\n", Entries);
	    else
		printf("%d New Items Loaded, use 'select all' to see them\n", Entries - oldEntries);
	}
	lmessage_overide = 0;
    }

    changed = 0;
    flock(fd, LOCK_UN);
    pop_break();
    return (1);
}

void
do_flock(int fd, int stat)
{
    if (flock(fd, stat | LOCK_NB) < 0) {
        puts ("File in use, Waiting for lock");
        flock (fd, stat);
        puts ("Have lock");
    }
}

int
load_file(int at, int pos)
{
    /*
     * Scan file
     */

    pos = (at) ? Entry[at].fpos : 0;

    while (position_at_from(pos)) {
	struct ENTRY *en;
	int l;

	if (at >= MaxEntries - 1) {
	    MaxEntries = (MaxEntries * 2 + 1000);
	    Entry = realloc(Entry, MaxEntries * sizeof(struct ENTRY));
	}
	en = &Entry[at];
	en->fpos = pos;
	en->from = FileData + pos + 5;

	{
	    int j;

            for (j = 0; Find[j].search != NULL; ++j)
		en->fields[j] = quo_quo;
	}

	/*
	 * Scan headers until blank line or next 'From '
	 */

	while ((l = position_next(&pos)) > 0) {
	    const char *p = FileData + pos;
	    int hl;

	    if (*p == '\n')
		break;
	    if (position_at_from(pos))
		break;

	    --l;	/* do not include newline */

	    /*
	     * Calculate header name.  Since header-continuances start 
	     * with a space, we don't have to worry about them.
	     */

	    for (hl = 0; hl < l && p[hl] != ':'; ++hl)
		;
	    if (hl < l)
		++hl;

	    /*
	     * Look for header in Find list and assign
	     * to fields as appropriate.
	     */

	    {
		int j;

		for (j = 0; Find[j].search != NULL; ++j) {
		    if (Find[j].len == hl && strncasecmp(Find[j].search, p, Find[j].len) == 0) {
			p = p + Find[j].len;
			while (p <  FileData + FileSize && (*p == ' ' || *p == '\t'))
			    ++p;
			en->fields[j] = p;
			break;
		    }
		}
	    }
	    /*
	     * skip to the newline terminating this line,
	     * makes position_next() a bit more efficient.
	     */
	    pos += l;
	}
	en->hlen = pos - en->fpos;

	while (!position_at_from(pos)) {
	    if (position_next(&pos) <= 0)
		break;
	}

	en->tlen = pos - en->fpos;

	/*
	 * bump at, figure out start position of next
	 * article.  Note that an Entry[] always exists
	 * for one past the last valid article containing
	 * the end-of-file position.
	 */

	if (++at > Entries) {
	    Entries = at;
	    Entry[at].fpos = pos;
	} else {
	    pos = Entry[at].fpos;
	}
    }
    return (1);
}

void
free_entry(void)
{
    if (Entry) {
	free_table(0);
	free(Entry);
    }
    Entry = calloc(sizeof(struct ENTRY), 1);
    MaxEntries = 1;

    if (FileData) {
	munmap((caddr_t)FileData, FileSize);
	FileSize = 0;
	FileData = NULL;
    }
    Listsize = 3;
    ChangesMade = 0;
    if (FileFd >= 0) {
	close(FileFd);
	FileFd = -1;
    }
}

/*
 * free_table() - only use to free table after end, or to 
 * free table at 0.
 */

int
free_table(int at)
{
    long pos = (at) ? Entry[at].fpos : 0;

    bzero(&Entry[at], (Entries - at) * sizeof(struct ENTRY));
    Entry[at].fpos = pos;
    Entries = at;
    return(pos);
}

int
save_file(int reload, int mark, int notmark)
{
    int fd;
    int count;
    int i;
    char scratch[64];

    for (i = 0; i < Entries; ++i) {
        if ((Entry[i].status & mark) != mark  ||
                (~Entry[i].status & notmark) != notmark) 
            break;
    }
    if (i == Entries && ChangesMade == 0) {
        m_select (Nulav, M_RESET, 0);
        puts ("No Changes Made");
        return (Entries);
    }
    if (FileFd <= 0)
        return (-1);

    /*
     * create scratch file
     */

    do_flock (FileFd, LOCK_EX);
    sprintf (scratch, "/var/tmp/dmail%d", (int)getpid());
    remove(scratch);
    fd = open(scratch, O_RDWR|O_CREAT|O_TRUNC|O_EXCL, MAILMODE);

    if (fd < 0) {
	flock(FileFd, LOCK_UN);
	printf("Unable to create scratch file: %s\n", scratch);
	return(-1);
    }

    /*
     * copy kept articles to a /tmp file.
     */

    count = 0;
    errno = 0;

    for (i = 0; i < Entries; ++i) {
        if ((Entry[i].status & mark) == mark  &&
	    (~Entry[i].status & notmark) == notmark
	) {
	    struct ENTRY *en = &Entry[i];
	    int pos;

            ++count;

	    pos = en->fpos;
	    position_next(&pos);
	    while (!position_at_from(pos)) {
		if (position_next(&pos) <= 0)
		    break;
	    }
	    if (write(fd, FileData + en->fpos, pos - en->fpos) != pos - en->fpos) {
		printf("\007ERROR writing %s, %s\n", scratch, strerror(errno));
		close(fd);
		remove(scratch);
		flock(FileFd, LOCK_UN);
		return(Entries);
	    }
        }
    }

    /*
     * If new mail has come in, append it to the scratch
     * file as well.
     */

    {
	struct stat st;

	if (fstat(FileFd, &st) == 0 && Entry[i].fpos < st.st_size) {
	    char buf[4096];
	    int n;

	    lseek(FileFd, Entry[i].fpos, 0);
	    while ((n = read(FileFd, buf, sizeof(buf))) > 0) {
		if (write(fd, buf, n) != n) {
		    printf("\007ERROR writing %s, %s\n", scratch, strerror(errno));
		    close(fd);
		    remove(scratch);
		    flock(FileFd, LOCK_UN);
		    return(Entries);
		}
	    }
	}
    }

    /*
     * rewrite mail file with scratch file
     */

    {
	char buf[4096];
	int n;

	lseek(FileFd, 0L, 0);
	lseek(fd, 0L, 0);
	while ((n = read(fd, buf, sizeof(buf))) > 0) {
	    write(FileFd, buf, n);
	}
	ftruncate(FileFd, lseek(FileFd, 0L, 1));
    }
    if (lseek(FileFd, 0L, 2) == 0 && !reload) {
        if (Did_cd == 0) {
            if (TruncateOnly == 0 && unlink(mail_file) == 0) 
                printf ("%s  Removed\n", mail_file);
            else
                printf ("0 messages left in %s\n", mail_file);
        }
    }
    close(fd);
    flock(FileFd, LOCK_UN);
    remove(scratch);
    free_entry();
    if (reload) {
	load_mail(0, 0);
    }
    m_select(Nulav, M_RESET, 0);
    return(count);
}

void
check_new_mail(void)
{
    struct stat st;

    push_break();

    if (FileFd >= 0 && fstat(FileFd, &st) == 0 && st.st_size != FileSize) {
	load_mail(Entries, 0);
    }

    pop_break();
}

int
write_file(char *file, int modes, int mark, int notmark)
{
    int i;
    int fd = -1;
    int notopen = 1;

    for (i = 0; i < Entries; ++i) {
        if ((Entry[i].status & mark) == mark  &&
	    (~Entry[i].status & notmark) == notmark
	) {
	    struct ENTRY *en = &Entry[i];
	    int pos;

            if (notopen) {
                notopen = 0;
                fd = open (file, O_APPEND | O_WRONLY | modes, MAILMODE);
                if (fd < 0)
                    return (-1);
                do_flock (fd, LOCK_EX);
            }

	    pos = en->fpos;
	    position_next(&pos);
	    while (!position_at_from(pos)) {
		if (position_next(&pos) <= 0)
		    break;
	    }
	    write(fd, FileData + en->fpos, pos - en->fpos);
        }
    }
    if (!notopen)
	close(fd);
    return (1);
}

int
write_prog(char *prog, int modes, int mark, int notmark)
{
    int i;
    FILE *fo = NULL;
    int notopen = 1;

    for (i = 0; i < Entries; ++i) {
        if ((Entry[i].status & mark) == mark  &&
	    (~Entry[i].status & notmark) == notmark
	) {
	    struct ENTRY *en = &Entry[i];
	    int pos;

            if (notopen) {
                notopen = 0;
                fo = popen (prog, "w");
                if (fo == NULL)
                    return (-1);
            }

	    pos = en->fpos;
	    position_next(&pos);
	    while (!position_at_from(pos)) {
		if (position_next(&pos) <= 0)
		    break;
	    }
	    fwrite(FileData + en->fpos, pos - en->fpos, 1, fo);
        }
    }
    if (!notopen)
	fclose(fo);
    return (1);
}

/*
 * Basic scheme: Each entry has a fields list.  Each entry in the fields list
 * is guarenteed to be a valid malloc'd pointer (except some may be set to
 * quo_quo).
 *
 * The find[] struct array holds the field name and length, the index 
 * corresponding to the index into the field[] in an Entry.
 *
 * The header and width arrays hold the list format.
 *
 * Return -1 on failure, -2 if 'Body' selected
 */

int
get_extra_ovr(char *str)
{
    register int i;

    if (strcasecmp(str, "body") == 0)
	return(-2);

    i = get_extra (str);
    if (i < 0) {
        i = add_extra (str);
        load_changes();
    }
    return (i);
}


/*
 * If there's room to add it, append to end.  
 * Else Find oldest field which doesn't exist in the setlist and replace it
 *  with the new one.
 */

int
add_extra(char *str)
{
    register int i, j, j_age, k;

    for (i = EXSTART; i < MAXTYPE; ++i) {
        if (Find[i].search == NULL)
            break;
        ++Find[i].age;
    }
    if (i == MAXTYPE) {                 /* No room to add onto end */
        j = j_age = -1;
        for (i = EXSTART; i < MAXTYPE; ++i) {
            for (k = 0; k < Listsize; ++k) {
                if (i == header[k])
                    break;
            }
            if (k == Listsize  &&  Find[i].age > j_age) {
                j = i;
                j_age = Find[i].age;
            }
        }
        i = j;
    }
    if (i < 0)
        return (-1);
    push_break();
    if (Find[i].search != NULL) 
        xfree (Find[i].search);
    Find[i].len = strlen(str);  
    Find[i].search = malloc (Find[i].len + 1);
    Find[i].notnew = Find[i].age = 0;
    strcpy (Find[i].search, str);
    changed = 1;
    for (j = 0; j < Entries; ++j) {
        Entry[j].fields[i] = quo_quo;
    }
    pop_break();
    return (i);
}

int
get_extra(char *str)
{
    int i;

    for (i = 0; Find[i].search; ++i) {
        if (strncasecmp(str, Find[i].search, strlen(str)) == 0) {
            Find[i].age = 0;
            return (i);
        }
    }
    return (-1);
}

int
m_select(char *sav[], int mode, int tag)
{
    char *ptr;
    char l_map[256];
    int idx[MAXLIST], ix = 0;
    int ok, not = 0, len, scr;
    register int i, j, avi;

    if (tag < 0) {
	tag = -tag;
	not = 1;
    }
        
    for (i = 0;i < 256; ++i)
        l_map[i] = i;
    for (i = 'A'; i <= 'Z'; ++i)
        l_map[i] += 'a' - 'A';
    hold_load();
    i = 0;
    idx[ix++] = get_extra_ovr (sav[i++]);
    for (; sav[i]; ++i) {
        if (strcmp (sav[i], ",") == 0  &&  sav[i + 1])
            idx[ix++] = get_extra_ovr (sav[++i]);
    }
    idx[ix] = -1;
    nohold_load();
    j = 1;
    push_break();
    for (i = 0; i < Entries; ++i) {
        if (mode == M_CONT  &&  Entry[i].no == 0)
            continue;

	ix = 0;
	ok = 0;
	avi = 1;

	/*
	 * Set ok based on tag match
	 */

	if (tag > 0) {
	    if (Entry[i].status & (1 << tag))
		ok = 1;
	} 

	/*
	 * Additional selection
	 */

        while ((ptr = sav[avi]) != NULL) {
	    const char *dest;
	    int destLen;

            if (ptr[0] == ','  &&  ptr[1] == '\0' && sav[avi+1]) {
                ++ix;
                avi += 2;
                continue;
            }
            if ((not = (*ptr == '!')) != 0)
                ++ptr;
            len = strlen (ptr);

	    if (idx[ix] == -2) {
		dest = FileData + Entry[i].fpos;
		destLen = Entry[i].tlen;
	    } else {
		dest = Entry[i].fields[idx[ix]];
		destLen = data_line_length(dest, 1);
	    }

            if (*ptr == '\0') {
                ok = 1;
                goto gotit;
            }
            while (destLen > 0 && *dest) {
                scr = 0;
                while (scr < destLen && l_map[(int)dest[scr]] == l_map[(int)ptr[scr]] && ptr[scr])
                    ++scr;
                if (ptr[scr] == '\0') {
                    ok = 1;
                    goto gotit;
                }
                ++dest;
		--destLen;
            }
            ++avi;
        }
gotit:
        Entry[i].no = (ok ^ not) ? j++ : 0;
    }
    pop_break();
    if (Current < 0)
        Current = 0;
    if (Entries) {
        if (Entry[Current].no == 0) {
            Current = indexof (1);
            if (Current < 0) {
                 Current = 0;
                 return (-1);
            }
        }   
    } else {
        Current = -1;
    }
    return (1);
}

int
position_at_from(int pos)
{
    if (pos <= FileSize - 5 && FileData[pos] == 'F' && strncmp(FileData + pos, "From ", 5) == 0)
	return(1);
    return(0);
}

/*
 * set *ppos to next line start and return the
 * length of the line inclusive of the newline.
 */

int
position_next(int *ppos)
{
    int pos = *ppos;
    int r = 0;

    while (pos < FileSize && FileData[pos] != '\n')
	++pos;
    if (pos < FileSize) {
	++pos;
	r = pos;
	while (r < FileSize && FileData[r] != '\n')
	    ++r;
	if (r < FileSize)
	    ++r;
	r -= pos;
    }
    *ppos = pos;
    return(r);
}

int
data_line_length(const char *p, int isHeader)
{
    if (p >= FileData && p < FileData + FileSize) {
	int b = p - FileData;
	int l;

	for (l = b; l < FileSize; ++l) {
	    if (FileData[l] == '\n') {
		if (isHeader &&
		    l + 1 < FileSize && 
		    (FileData[l+1] == ' ' || FileData[l+1] == '\t')) {
		    continue;
		}
		break;
	    }
	}
	if (l < FileSize)
	    ++l;
	return(l - b);
    }
    return(strlen(p));
}

